(*----------------------------------------------------------------------------
* APPLY_PATTERN
*-----------------------------------------------------------------------------
* DESCRIPTION : Applies pattern on a single pixel
*
* TYPE : Function. Returns an int.
*
* ARGUMENTS :
* matsrc : Source Matrix to be treated
* matftr : Pattern Matrix required to apply filter
* x      : x coordinate of the pixel that will be treated
* y      : y coordinate of the pixel that will be treated
-----------------------------------------------------------------------------*)


(* Collisions management *)
let apply_pattern matsrc matftr x y =
    (* Width of the matrix  *)
    let ftrsize = Array.length matftr in

    (* Sum of all values *)
    let sum = ref 0 in

    (* Increments  *)
    let increment_x = ref 0 in
    let increment_y = ref 0 in

    (* Sum of all coeficients *)
    let coef_sum = ref 0 in

    begin

        (* Browsing width *)
        for i = x - (ftrsize/2) to x + (ftrsize/2) do
            (
                (* Browsing height *)
                for j = (y - (ftrsize/2)) to (y + (ftrsize/2)) do
                    (
                        (* Multiplies box's value by the coefficient *)
                        sum := !sum + matsrc.(i).(j) *
                        matftr.(!increment_x).(!increment_y);

                        (* Sum of the filter's coefficient  *)
                        coef_sum := !coef_sum +
                        matftr.(!increment_x).(!increment_y);

                        increment_y := !increment_y + 1 ;
                    );
                done;
                increment_y := 0;
                increment_x := !increment_x + 1 ;
             )
        done;

        !sum / !coef_sum;
    end

(*-----------------------------------------------------------------------------
* PROCESS
* -----------------------------------------------------------------------------
* DESCRIPTION : Processing all pixels
*
* TYPE : Function, returns matout, int matrix
*
* ARGUMENTS :
* matsrc : Source Matrix to be treated
* matftr : Pattern Matrix required to apply treatments (Filter)
* ---------------------------------------------------------------------------*)

let process matsrc matftr =

    (* Sizes of the matrix *)
    let (x,y) = (Array.length matsrc, Array.length matsrc.(0)) in
    let ftrsize = Array.length matftr in

    (* Output matrix *)
    let matout = matsrc in

    begin

        (* Mapping width  *)
        for i = (ftrsize / 2) to (x - (ftrsize / 2)) - 1 do

            (* Mapping height *)
            for j = (ftrsize / 2) to (y - (ftrsize / 2)) - 1 do
                matout.(i).(j) <- apply_pattern matsrc matftr i j;
            done;

        done;

        matout;
    end


(*----------------------------------------------------------------------------
* NOISE_REDUCTION
*-----------------------------------------------------------------------------
* DESCRIPTION : Defines the value of a single pixel for the noise reduction
*
* TYPE : Function. Returns a new int matrix
*
* ARGUMENTS :
* matsrc   : Source Matrix to be treated
* ftrsize  : Size representing the number of pixels to take into acount
* treshold : Number of black pixels to consider it's not noise
* x        : x coordinate of the pixel that will be treated
* y        : y coordinate of the pixel that will be treated
-----------------------------------------------------------------------------*)

let noise_reduction matsrc ftrsize treshold x y =
    (* Number of black pixels around *)
    let pxnbr = ref 0 in

    begin

        (* Browsing width *)
        for i = x - (ftrsize/2) to x + (ftrsize/2) do
            (
                (* Browsing height *)
                for j = (y - (ftrsize/2)) to (y + (ftrsize/2)) do
                    (
                        (* If the pixel is black *)
                        if matsrc.(i).(j) = 0 then
                            pxnbr := !pxnbr + 1;
                    );
                done;
             )
        done;

        let condition = (!pxnbr < treshold) in
        if condition then 255
        else 0;
    end

(*----------------------------------------------------------------------------
* REDUCE_NOISE
*-----------------------------------------------------------------------------
* DESCRIPTION : Deletes little noises on the matrix by counting the number of
* black pixels around the pixel.
*
* TYPE : Function. Returns a new int matrix
*
* ARGUMENTS :
* matsrc   : Source Matrix to be treated
* ftrsize  : Size representing the number of pixels to take into acount
* treshold : Number of black pixels to consider it's not noise
*
* COMMENTS : Should be used on a black & white matrix
-----------------------------------------------------------------------------*)

let reduce_noise matsrc treshold ftrsize =

    (* Sizes of the matrix *)
    let (x,y) = (Array.length matsrc, Array.length matsrc.(0)) in

    (* Output matrix *)
    let matout = matsrc in

    begin

        (* Mapping width *)
        for i = (ftrsize / 2) to (x - (ftrsize / 2)) - 1 do

            (* Mapping height *)
            for j = (ftrsize / 2) to (y - (ftrsize / 2)) - 1 do

                (* If the pixel is black *)
                if matsrc.(i).(j) = 0 then
                matout.(i).(j) <- noise_reduction matsrc ftrsize treshold i j;
            done;

        done;

        matout;
    end


(*-----------------------------------------------------------------------------
* FILTERS
* ---------------------------------------------------------------------------*)

let grey = [|[|1|]|]

let moyenne = Array.make 3  (Array.make 3 1)

let moyenne_pond = [| [|1;2;1|] ;
                      [|2;4;2|] ;
                      [|1;2;1|] |]

let moyenne_pond2 = [| [|2;5;2|] ;
                       [|5;20;5|] ;
                       [|2;5;2|] |]

(*-----------------------------------------------------------------------------
* MAIN
* -----------------------------------------------------------------------------
* DESCRIPTION : main call to apply a filter to an image
*
* TYPE : Function, returns the filtered matrix
*
* ARGUMENTS :
* matsrc : Source Matrix to be treated
* filter : value that specifies the filter to be applied
* ---------------------------------------------------------------------------*)

let main matsrc filter =
    begin
        let matsrc_1coord    = Matrix.to1coord matsrc in
        let matsrc_filtered  = process matsrc_1coord filter in
        let matsrc_finished  = Matrix.to3coord matsrc_filtered in
        matsrc_finished;
    end

(*-----------------------------------------------------------------------------
* MAIN_NOISE
* -----------------------------------------------------------------------------
* DESCRIPTION : main call to apply noise reduction to an image
*
* TYPE : Function, returns the filtered matrix
*
* ARGUMENTS :
* matsrc   : Source Matrix to be treated
* size     : Size of the zone to be treated (3 for 3x3, 5 for 5x5..)
* treshold : Number of pixel to consider it's not noise
* ---------------------------------------------------------------------------*)

let main_noise matsrc =
    begin
        let matsrc_1coord     = Matrix.to1coord matsrc in
        let matsrc_filtered   = reduce_noise matsrc_1coord 2 3 in
	let matsrc_filtered_2 = reduce_noise matsrc_filtered 4 9 in
        let matsrc_finished   = Matrix.to3coord matsrc_filtered_2 in
        matsrc_finished;
    end



